#version 330 core

#info See blog post: <https://mathr.co.uk/blog/2021-11-12_atom_domains_and_newton_basins.html>

#include "TwoD.frag"

#group Mandelbrot

// period of nuclei to find
uniform int Period; slider[1,1,64]
// number of iterations for exterior DE colouring
uniform int Iterations; slider[1,64,4096]
// number of iterations for Newton's method
uniform int NewtonSteps; slider[1,64,256]
// use reduced polynomials for Newton's method
uniform bool NewtonReduced; checkbox[true]

// http://lolengine.net/blog/2013/07/27/rgb-to-hsv-in-glsl
vec3 hsv2rgb(vec3 c)
{
  vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
  vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
  return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

// find periodic nucleus using Newton's method
Complexf Mandelbrot_nucleus(Complexf c0, int period, int steps, bool reduced)
{
  Complexf c = c0;
  for (int j = 0; j < steps; ++j)
  {
    Dual1cf G = dual1cf(complexf(1.0, 0.0));
    Dual1cf z = dual1cf(complexf(0.0, 0.0));
    for (int l = 1; l <= period; ++l)
    {
      z = add(sqr(z), dual1cf(c, 0));
      if (l < period && period % l == 0 && reduced)
        G = mul(z, G);
    }
    G = div(z, G);
    c = sub(c, div(G.x,  G.d[0]));
  }
  return c;
}

// find exterior distance estimate
vec4 Mandelbrot_exterior_de(Complexf c0, int n)
{
  Dual1cf z = dual1cf(complexf(0.0, 0.0));
  Dual1cf c = dual1cf(c0, 0);
  float mz2 = 1.0 / 0.0;
  int period = 0;
  int i = 0;
  for (i = 1; i <= n; ++i)
  {
    z = add(sqr(z), c);
    if (norm(z.x) > 256.0)
      break;
    if (norm(z.x) < mz2 && i <= Period)
    {
      mz2 = norm(z.x);
      period = i;
    }
  }
  if (norm(z.x) > 256.0)
  {
    return vec4(length(z.x) * log(length(z.x)) / length(z.d[0]), float(period), float(i), arg(z.x));
  }
  return vec4(-1.0, float(period), float(i), 0.0);
}

// main entry point
vec3 color(vec2 p, vec2 dx, vec2 dy)
{
  Complexf z0 = complexf(p.x, p.y);
  Complexf x = complexf(dx.x, dx.y);
  Complexf y = complexf(dy.x, dy.y);
  Complexf z00 = Mandelbrot_nucleus(z0, Period, NewtonSteps, NewtonReduced);
  Complexf z01 = Mandelbrot_nucleus(add(z0, x), Period, NewtonSteps, NewtonReduced);
  Complexf z10 = Mandelbrot_nucleus(add(z0, y), Period, NewtonSteps, NewtonReduced);
  Complexf z11 = Mandelbrot_nucleus(add(z0, add(x, y)), Period, NewtonSteps, NewtonReduced);

  float q = length(sub(z11, z00)) / length(add(x, y)) + length(sub(z01, z10)) / length(add(x, y));

  vec4 de00 = Mandelbrot_exterior_de(z0, Iterations);
  vec4 de01 = Mandelbrot_exterior_de(add(z0, x), Iterations);
  vec4 de10 = Mandelbrot_exterior_de(add(z0, y), Iterations);
  vec4 de11 = Mandelbrot_exterior_de(add(z0, add(x, y)), Iterations);

  bool gridy = de00.z != de11.z || de01.z != de10.z;
  bool gridx = (de00.w > 0.0) != (de11.w > 0.0) || (de01.w > 0.0) != (de10.w > 0.0);
  bool grid = gridx || gridy;

  vec4 de = de00;

  // saturated near nucleus locations
  float hue = 0.0;
  float sat = de.y == float(Period) ? 0.5 : 0.0;
  float val = de.y == float(Period) ? 2.0 : 1.0;

  // dark Mandelbrot set boundary
  float b = de.x / length(vec4(dx, dy));
  val *= clamp(4.0 * b, 0.0, 0.5);
  if (grid) val *= 0.5;

  // white Newton basin boundaries
  if (! (q < 1.0)) return vec3(1.0);

  // bad value checks
  if (!(z00 == z00)) return vec3(1.0);
  if (isinf(de.x) || isnan(de.x)) return vec3(1.0);
  if (isinf(b) || isnan(b)) return vec3(1.0);

  return hsv2rgb(vec3(hue, sat, val));
}

#preset Default
Zoom = 400
Center = -0.8173,0.1899
Jitter = 1
Samples = 32
Period = 18
Iterations = 1000
NewtonReduced = true
NewtonSteps = 64
#endpreset

#preset Range-0001-1800
ZoomFactor1:OutQuad:2:-9:0:1:1800:0.3:1:1.7:1:0
#endpreset
